/* MAA - Message Authentication Algorithm - ISO 8731-2
 *
 * C implementation by Niels Mller <nisse@lysator.liu.se> 1996
 */

#include <ISO_MAA.h>
#include <stdio.h>

/* 64 bit integer */
#define DWORD unsigned long long

/* Constants */
#define MAA_A 0x02040801
#define MAA_B 0x00804021
#define MAA_C 0xbfef7fdf
#define MAA_D 0x7dfefbff

/* Primitive operations */
static inline MAA_WORD
CYC(MAA_WORD x)   /* Cyclic shift one bit left */
{ 
  return (x << 1) | (x>>31);
}

#define AND(x,y) ((x) & (y))
#define OR(x,y) ((x) | (y))
#define XOR(x,y) ((x) ^ (y))

/* Addition mod 2^32 */
#define ADD(x,y) ((x) + (y))

static inline MAA_WORD
CAR(MAA_WORD x, MAA_WORD y)  /* Carry when adding x and y */
{
  return x > (0xffffffff - y);
}

static inline DWORD
MUL64(MAA_WORD x, MAA_WORD y)  /* 32*32 -> 64 bit addition */
{
  DWORD z = (DWORD) x * y;
  return z;
}

static inline MAA_WORD
MUL1(MAA_WORD x, MAA_WORD y)   /* Result = x * y (mod 2^32-1) */
{
  DWORD z;
  MAA_WORD l, u;

  z = MUL64(x,y);
  u = z >> 32; l = z & 0xffffffff;
  return ADD(ADD(u,l), CAR(u,l));
}

static inline MAA_WORD
MUL2(MAA_WORD x, MAA_WORD y)  /* Result = x * y (mod 2^32 - 2) */
{
  DWORD z;
  MAA_WORD l, u, f;
  
  z = MUL64(x,y);
  u = z >> 32; l = z & 0xffffffff;

  f = ADD(ADD(u,u), 2*CAR(u,u));
  return ADD(ADD(f,l), 2*CAR(f,l));
}

static inline MAA_WORD
MUL2A(MAA_WORD x, MAA_WORD y) /* Alternative to MUL2, that works if the
		               * most significant bit of at least one
		               * of the arguments is 0. */
{
  DWORD z;
  MAA_WORD l, u, d;

  z = MUL64(x,y);
  u = z >> 32; l = z & 0xffffffff;
  d = ADD(u,u);
  return ADD(ADD(d,l), 2*CAR(d,l));
}

static inline unsigned char
OCTET(unsigned i, MAA_WORD x) /* Extract the i:th most significant byte */
{
  return (x >> (24 - 8*i)) & 0xff;
}

static inline MAA_WORD
ASSEMBLE_OCTETS(unsigned char c[4]) /* Put 4 octets together to a word */
{
  unsigned i;
  MAA_WORD x;
  
  for (i = 0, x = 0; i<4; i++)
    x = (x<<8) + c[i];
  return x;
}

static unsigned char
BYTPAT(MAA_WORD x, MAA_WORD y,
       MAA_WORD *x1, MAA_WORD *y1)   /* Avoid strings of ones and zeros */
{
  unsigned i;
  unsigned char b[8];
  unsigned char p;

  /* Convert to octets */
  for (i=0; i<4; i++)
    {
      b[i]   = OCTET(i, x);
      b[4+i] = OCTET(i, y);
    }

  for (p = 0, i = 0; i<8; i++)
    {
      p *= 2;
      switch (b[i])
	{
	case 0:
	  p++;
	  b[i] = p;
	  break;
	case 0xff:
	  p++;
	  b[i] = 0xff - p;
	  break;
	default:
	  ;
	}
    }

  *x1 = ASSEMBLE_OCTETS(b);
  *y1 = ASSEMBLE_OCTETS(b+4);
  return p;
}


/* Shorthands used below */
#define MUL3(z, x, y) do { \
  z[0] = MUL1(x[0],y[0]); z[1] = MUL2(x[1], y[1]); \
} while (0)

#define XOR2(h, j) ((h) = XOR((j)[0], (j)[1]))

static void
PRELUDE(struct MAA_key *key)
{
  MAA_WORD j1, k1;
  MAA_WORD j2[2], j4[2], j6[2], j8[2];
  MAA_WORD h4, h6, h8, h1, h5, h7, h9;
  MAA_WORD k2[2], k4[2], k5[2], k7[2], k9[2];
  unsigned char p;
  MAA_WORD q;
  
  p = BYTPAT(key->J, key->K, &j1, &k1);
  q = (1 + p) * (1 + p);

  j2[0] = MUL1(j1, j1); j2[1] = MUL2(j1, j1);
  MUL3(j4, j2, j2);
  MUL3(j6, j2, j4);
  MUL3(j8, j2, j6);

  XOR2(h4, j4);
  XOR2(h6, j6);
  XOR2(h8, j8);

  k2[0] = MUL1(k1, k1);    k2[1] = MUL2(k1, k1);
  MUL3(k4, k2, k2);
  k5[0] = MUL1(k1, k4[0]); k5[1] = MUL2(k1, k4[1]);
  MUL3(k7, k2, k5);
  MUL3(k9, k2, k7);

  XOR2(h1, k5);
  h5 = MUL2(h1, q);
  XOR2(h7, k7);
  XOR2(h9, k9);

  BYTPAT(h4, h5, &(key->X0), &(key->Y0));
  BYTPAT(h6, h7, &(key->V0), &(key->W));
  BYTPAT(h8, h9, &(key->S), &(key->T));
}

void
MAA_key(struct MAA_key *key, MAA_WORD j, MAA_WORD k)
{
  key->J = j; key->K = k;
  PRELUDE(key);
}

inline void
MAA_reset(struct MAA_state *state)
{
  state->X = state->key->X0;
  state->Y = state->key->Y0;
  state->V = state->key->V0;
  state->block_count = state->buffer_count = 0;
}

static inline void
MAA_loop(struct MAA_state *state, MAA_WORD m)
{
  MAA_WORD e, f, g;
  
  state->V = CYC(state->V);
  e = XOR(state->V, state->key->W);
  state->X = XOR(state->X, m);  state->Y = XOR(state->Y, m);
  f = AND(MAA_C, OR(MAA_A, ADD(e, state->Y)));
  g = AND(MAA_D, OR(MAA_B, ADD(e, state->X)));

  state->X = MUL1(state->X, f); state->Y = MUL2A(state->Y, g);
}

static MAA_WORD
MAA_coda(struct MAA_state *state)
{
  MAA_loop(state, state->key->S);
  MAA_loop(state, state->key->T);
  return XOR(state->X, state->Y);
}

static inline void
MAA_process_block(struct MAA_state *state, MAA_WORD m)
{
  if (state->block_count == 256)
    {
      MAA_WORD z = MAA_coda(state);
      MAA_reset(state);
      MAA_loop(state, z);
      state->block_count = 0;
    }
  MAA_loop(state, m);
  state->block_count++;
}

void
MAA_update(struct MAA_state *state,
	   unsigned char *message,
	   unsigned length)
{
  while ((length > 0) && (state->buffer_count < 4))
    {
      state->buffer[state->buffer_count++] = *message++;
      length--;
    }
  if (state->buffer_count == 4)
    {
      MAA_process_block(state, ASSEMBLE_OCTETS(state->buffer));
      for ( ; length >= 4; length -= 4, message += 4)
	MAA_process_block(state, ASSEMBLE_OCTETS(message));
      for (state->buffer_count = 0; length; length--)
	state->buffer[state->buffer_count++] = *message++;
    }
}

MAA_WORD
MAA_mac(struct MAA_state *state)
{
  MAA_WORD z;
  if (state->buffer_count)
    {
      fprintf(stderr, "MAA: Not an even number of blocks,\n"
	      "and no padding method defined.\n");
      abort();
    }
  z = MAA_coda(state);
  MAA_reset(state);
  return z;
}

#ifdef MAA_TEST

void
test_prelude(struct MAA_key *key, MAA_WORD x0, MAA_WORD y0,
	     MAA_WORD v0, MAA_WORD w, MAA_WORD s, MAA_WORD t)
{
  PRELUDE(key);
  if ((x0 != key->X0)
      || (y0 != key->Y0)
      || (v0 != key->V0)
      || (w != key->W)
      || (s != key->S)
      || (t != key->T))
    {
      printf("Prelude failed!\n"
	     "Expected: [%8x, %8x, %8x, %8x, %8x, %8x]\n"
	     "Got:      [%8x, %8x, %8x, %8x, %8x, %8x]\n",
	     x0, y0, v0, w, s, t,
	     key->X0, key->Y0, key->V0, key->W, key->S, key->S);
    }
}

void
test_round(struct MAA_state *state, MAA_WORD block,
	   MAA_WORD x, MAA_WORD y)
{
  MAA_process_block(state, block);
  if ((x != state->X) || (y != state->Y))
    {
      printf("Round failed!\n"
	     "Expected: [%8x, %8x]\n"
	     "Got:      [%8x, %8x]\n",
	     x, y, state->X, state->Y);
    }
}

MAA_WORD test_output[20][2] =
{
  {0x303ff4aa, 0x1277a6d4},
  {0x55dd063f, 0x4c49aae0},
  {0x51af3c1d, 0x5bc02502},
  {0xa44aaac0, 0x63c70dba},
  {0x4d53901a, 0x2e80ac30},
  {0x5f38eef1, 0x2a6091ae},
  {0xf0239dd5, 0x3dd81ac6},
  {0xeb35b97f, 0x9372cdc6},
  {0x4da124a1, 0xc6b1317e},
  {0x7f839576, 0x74b39176},
  {0x11a9d254, 0xd78634bc},
  {0xd8804ca5, 0xfdc1a8ba},
  {0x3f6f7248, 0x11ac46b8},
  {0xacbc13dd, 0x33d5a466},
  {0x4ce933e1, 0xc21a1846},
  {0xc1ed90dd, 0xcd959b46},
  {0x3cd54deb, 0x613f8e2a},
  {0xbba57835, 0x07c72eaa},
  {0xd7843fdc, 0x6ad6e8a4},
  {0x5eba06c2, 0x91896cfa}
};

int main(int argc, char ** argv)
{
  struct MAA_key key;
  struct MAA_state state = { key: &key };
  MAA_WORD mac;
  unsigned i;
  
  key.J = 0x00ff00ff; key.K = 0;
  test_prelude(&key,
	       0x4a645a01, 0x50dec930,
	       0x5cca3239, 0xfeccaa6e,
	       0x51ede9c7, 0x24b66fb5);
  MAA_reset(&state);
  test_round(&state, 0x55555555, 0x48b204d6, 0x5834a585);
  test_round(&state, 0xaaaaaaaa, 0x4f998e01, 0xbe9f0917);
  
  mac = MAA_mac(&state);
  printf("MAC: %8x\n", mac);
  if (mac != 0xf14d6e28)
    printf("Failed!\n");

  test_round(&state, 0xaaaaaaaa, 0x6aebacf8, 0x9db15cf6);
  test_round(&state, 0x55555555, 0x270eedaf, 0xb8142629);

  mac = MAA_mac(&state);
  printf("MAC: %8x\n", mac);
  if (mac != 0xa93bd410)
    printf("Failed!\n");

  key.J = 0x55555555; key.K = 0x5a35d667;
  test_prelude(&key, 0x34acf886, 0x7397c9ae,
	       0x7201f4dc, 0x2829040b,
	       0x9e2e7b36, 0x13647149);
  MAA_reset(&state);
  
  test_round(&state, 0, 0x2fd76ffb, 0x550d91ce);
  test_round(&state, 0xffffffff, 0xa70fc148, 0x1d10d8d3);
  
  mac = MAA_mac(&state);
  printf("MAC: %8x\n", mac);
  if (mac != 0xb99a62de)
    printf("Failed!\n");

  test_round(&state, 0xffffffff, 0x8dc8bbde, 0xfe4e5bdd);
  test_round(&state, 0,          0xcbc865ba, 0x0297af6f);

  mac = MAA_mac(&state);
  printf("MAC: %8x\n", mac);
  if (mac != 0xa018c83b)
    printf("Failed!\n");

  MAA_key(&key, 0x80018001, 0x80018000); MAA_reset(&state);
  for (i=0; i < 20; i++)
    test_round(&state, 0, test_output[i][0], test_output[i][1]);

  mac = MAA_mac(&state);
  printf("MAC: %8x\n", mac);
  if (mac != 0xdb79fbdc)
    printf("Failed!\n");
}
#endif MAA_TEST
